@@ -44,16 +44,16 @@
 #include <processor.h>
 #include <misc.h>
 #include <uuid.h>
-#include <multiboot.h>
+#include <loader.h>
 #include <hash.h>
 #include <tb_error.h>
 #define PRINT printk
-#include <tb_policy.h>
 #include <mle.h>
-#include <tpm.h>
 #include <loader.h>
 #include <tboot.h>
 #include <integrity.h>
+#include <tpm.h>
+#include <tb_policy.h>
 #include <lcp2.h>
 #include <lcp_hlp.h>
 #include <cmdline.h>
@@ -110,6 +110,7 @@ static const tb_policy_map_t g_policy_map[] = {
     { TB_POLTYPE_CONT_VERIFY_FAIL,             TB_POLACT_HALT,
       {
           {TB_ERR_MODULE_VERIFICATION_FAILED,  TB_POLACT_CONTINUE},
+          {TB_ERR_NV_VERIFICATION_FAILED,      TB_POLACT_CONTINUE},
           {TB_ERR_POLICY_NOT_PRESENT,          TB_POLACT_CONTINUE},
           {TB_ERR_POLICY_INVALID,              TB_POLACT_CONTINUE},
           {TB_ERR_NONE,                        TB_POLACT_CONTINUE},
@@ -134,8 +135,9 @@ static uint8_t _policy_index_buf[MAX_POLICY_SIZE];
 static const tb_policy_t _def_policy = {
     version        : 2,
     policy_type    : TB_POLTYPE_CONT_NON_FATAL,
+    hash_alg       : TB_HALG_SHA1,
     policy_control : TB_POLCTL_EXTEND_PCR17,
-    num_entries    : 2,
+    num_entries    : 3,
     entries        : {
         {   /* mod 0 is extended to PCR 18 by default, so don't re-extend it */
             mod_num    : 0,
@@ -148,6 +150,13 @@ static const tb_policy_t _def_policy = {
             pcr        : 19,
             hash_type  : TB_HTYPE_ANY,
             num_hashes : 0
+        },
+        {   /* NV index for geo-tagging will be extended to PCR 22 */
+            mod_num    : TB_POL_MOD_NUM_NV_RAW,
+            pcr        : 22,
+            hash_type  : TB_HTYPE_ANY,
+            nv_index   : 0x40000010,
+            num_hashes : 0
         }
     }
 };
@@ -156,8 +165,9 @@ static const tb_policy_t _def_policy = {
 static const tb_policy_t _def_policy_da = {
     version        : 2,
     policy_type    : TB_POLTYPE_CONT_NON_FATAL,
+    hash_alg       : TB_HALG_SHA1,
     policy_control : TB_POLCTL_EXTEND_PCR17,
-    num_entries    : 2,
+    num_entries    : 3,
     entries        : {
         {   /* mod 0 is extended to PCR 17 by default, so don't re-extend it */
             mod_num    : 0,
@@ -170,6 +180,13 @@ static const tb_policy_t _def_policy_da = {
             pcr        : 17,
             hash_type  : TB_HTYPE_ANY,
             num_hashes : 0
+        },
+        {   /* NV index for geo-tagging will be extended to PCR 22 */
+            mod_num    : TB_POL_MOD_NUM_NV_RAW,
+            pcr        : 22,
+            hash_type  : TB_HTYPE_ANY,
+            nv_index   : 0x40000010,
+            num_hashes : 0
         }
     }
 };
@@ -184,8 +201,9 @@ static const tb_policy_t* g_policy = &_def_policy;
  *
  * policy_index_size is in/out
  */
-static bool read_policy_from_tpm(tpm_nv_index_t index,
-                void* policy_index, size_t *policy_index_size)
+static bool 
+read_policy_from_tpm(uint32_t index,
+                     void* policy_index, size_t *policy_index_size)
 {
 #define NV_READ_SEG_SIZE    256
     unsigned int offset = 0;
@@ -193,16 +211,16 @@ static bool read_policy_from_tpm(tpm_nv_index_t index,
     uint32_t ret, index_size;
 
     if ( policy_index_size == NULL ) {
-        printk("size is NULL\n");
+        printk(TBOOT_ERR"size is NULL\n");
         return false;
     }
 
-    ret = tpm_get_nvindex_size(0, index, &index_size);
-    if ( ret != TPM_SUCCESS )
+    ret = g_tpm->get_nvindex_size(g_tpm, 0, index, &index_size);
+    if ( !ret )
         return false;
 
     if ( index_size > *policy_index_size ) {
-        printk("policy in TPM NV %x was too big for buffer\n", index);
+        printk(TBOOT_WARN"policy in TPM NV %x was too big for buffer\n", index);
         index_size = *policy_index_size;
     }
 
@@ -215,17 +233,17 @@ static bool read_policy_from_tpm(tpm_nv_index_t index,
             data_size = (uint32_t)(index_size - offset);
 
         /* read! */
-        ret = tpm_nv_read_value(0, index, offset,
-                                (uint8_t *)policy_index + offset, &data_size);
-        if ( ret != TPM_SUCCESS || data_size == 0 )
+        ret = g_tpm->nv_read(g_tpm, 0, index, offset,
+                             (uint8_t *)policy_index + offset, &data_size);
+        if ( !ret || data_size == 0 )
             break;
 
         /* adjust offset */
         offset += data_size;
     } while ( offset < index_size );
 
-    if ( offset == 0 && ret != TPM_SUCCESS ) {
-        printk("Error: read TPM error: 0x%x from index %x.\n", ret, index);
+    if ( offset == 0 && !ret ) {
+        printk(TBOOT_ERR"Error: read TPM error: 0x%x from index %x.\n", ret, index);
         return false;
     }
 
@@ -245,6 +263,9 @@ static bool unwrap_lcp_policy(void)
     void* lcp_base;
     uint32_t lcp_size;
 
+    // scaffolding
+    printk(TBOOT_INFO"in unwrap_lcp_policy\n");
+
     if ( txt_is_launched() ) {
         txt_heap_t *txt_heap = get_txt_heap();
         os_sinit_data_t *os_sinit_data = get_os_sinit_data_start(txt_heap);
@@ -253,8 +274,8 @@ static bool unwrap_lcp_policy(void)
         lcp_size = (uint32_t)os_sinit_data->lcp_po_size;
     }
     else {
-        extern multiboot_info_t *g_mbi;
-        if ( !find_lcp_module(g_mbi, &lcp_base, &lcp_size) )
+        extern loader_ctx *g_ldr_ctx;
+        if ( !find_lcp_module(g_ldr_ctx, &lcp_base, &lcp_size) )
             return false;
     }
 
@@ -306,26 +327,26 @@ tb_error_t set_policy(void)
 {
     /* try to read tboot policy from TB_POLICY_INDEX in TPM NV */
     size_t policy_index_size = sizeof(_policy_index_buf);
-    printk("reading Verified Launch Policy from TPM NV...\n");
-    if ( read_policy_from_tpm(TB_POLICY_INDEX,
+    printk(TBOOT_INFO"reading Verified Launch Policy from TPM NV...\n");
+    if ( read_policy_from_tpm(g_tpm->tb_policy_index,
              _policy_index_buf, &policy_index_size) ) {
-        printk("\t:%lu bytes read\n", policy_index_size);
+        printk(TBOOT_DETA"\t:%lu bytes read\n", policy_index_size);
         if ( verify_policy((tb_policy_t *)_policy_index_buf,
                  policy_index_size, true) ) {
             goto policy_found;
         }
     }
-    printk("\t:reading failed\n");
+    printk(TBOOT_WARN"\t:reading failed\n");
 
     /* tboot policy not found in TB_POLICY_INDEX, so see if it is wrapped
      * in a custom element in the PO policy; if so, SINIT will have verified
      * the policy and policy data for us; we just need to ensure the policy
      * type is LCP_POLTYPE_LIST (since we could have been give a policy data
      * file even though the policy was not a LIST */
-    printk("reading Launch Control Policy from TPM NV...\n");
-    if ( read_policy_from_tpm(INDEX_LCP_OWN,
+    printk(TBOOT_INFO"reading Launch Control Policy from TPM NV...\n");
+    if ( read_policy_from_tpm(g_tpm->lcp_own_index,
              _policy_index_buf, &policy_index_size) ) {
-        printk("\t:%lu bytes read\n", policy_index_size);
+        printk(TBOOT_DETA"\t:%lu bytes read\n", policy_index_size);
         /* assume lcp policy has been verified by sinit already */
         lcp_policy_t *pol = (lcp_policy_t *)_policy_index_buf;
         if ( pol->policy_type == LCP_POLTYPE_LIST && unwrap_lcp_policy() ) {
@@ -335,10 +356,10 @@ tb_error_t set_policy(void)
                 goto policy_found;
         }
     }
-    printk("\t:reading failed\n");
+    printk(TBOOT_WARN"\t:reading failed\n");
 
     /* either no policy in TPM NV or policy is invalid, so use default */
-    printk("failed to read policy from TPM NV, using default\n");
+    printk(TBOOT_WARN"failed to read policy from TPM NV, using default\n");
     g_policy = g_using_da ? &_def_policy_da : &_def_policy;
     policy_index_size = calc_policy_size(g_policy);
 
@@ -349,15 +370,21 @@ tb_error_t set_policy(void)
         return TB_ERR_POLICY_NOT_PRESENT;
 
 policy_found:
+    /* compatible with tb_policy tools for TPM 1.2 */
+    {
+        tb_policy_t *tmp_policy = (tb_policy_t *)_policy_index_buf;
+        if (tmp_policy->hash_alg == 0)
+            tmp_policy->hash_alg = TB_HALG_SHA1;
+    }
     g_policy = (tb_policy_t *)_policy_index_buf;
     return TB_ERR_NONE;
 }
 
 /* hash current policy */
-bool hash_policy(tb_hash_t *hash, uint8_t hash_alg)
+bool hash_policy(tb_hash_t *hash, uint16_t hash_alg)
 {
     if ( hash == NULL ) {
-        printk("Error: input parameter is wrong.\n");
+        printk(TBOOT_ERR"Error: input parameter is wrong.\n");
         return false;
     }
 
@@ -366,12 +393,12 @@ bool hash_policy(tb_hash_t *hash, uint8_t hash_alg)
 }
 
 /* generate hash by hashing cmdline and module image */
-static bool hash_module(tb_hash_t *hash, uint8_t hash_alg,
+static bool hash_module(hash_list_t *hl,
                         const char* cmdline, void *base,
                         size_t size)
 {
-    if ( hash == NULL ) {
-        printk("Error: input parameter is wrong.\n");
+    if ( hl == NULL ) {
+        printk(TBOOT_ERR"Error: input parameter is wrong.\n");
         return false;
     }
 
@@ -385,27 +412,83 @@ static bool hash_module(tb_hash_t *hash, uint8_t hash_alg,
         cmdline = "";
     else
         cmdline = skip_filename(cmdline);
-    if ( !hash_buffer((const unsigned char *)cmdline, strlen(cmdline), hash,
-                      hash_alg) )
-        return false;
 
-    /* hash image and extend into cmdline hash */
-    tb_hash_t img_hash;
-    if ( !hash_buffer(base, size, &img_hash, hash_alg) )
-        return false;
-    if ( !extend_hash(hash, &img_hash, hash_alg) )
+    switch (g_tpm->extpol) {
+    case TB_EXTPOL_FIXED: 
+        hl->count = 1;
+        hl->entries[0].alg = g_tpm->cur_alg;
+
+        if ( !hash_buffer((const unsigned char *)cmdline, strlen(cmdline),
+                    &hl->entries[0].hash, g_tpm->cur_alg) )
+            return false;
+        /* hash image and extend into cmdline hash */
+        tb_hash_t img_hash;
+        if ( !hash_buffer(base, size, &img_hash, g_tpm->cur_alg) )
+            return false;
+        if ( !extend_hash(&hl->entries[0].hash, &img_hash, g_tpm->cur_alg) )
+            return false;
+
+        break;
+
+    case TB_EXTPOL_AGILE: 
+    {
+        hash_list_t img_hl;
+        if ( !g_tpm->hash(g_tpm, 2, (const unsigned char *)cmdline,
+                strlen(cmdline), hl) ) {
+            if ( !g_tpm->hash(g_tpm, 2, base, size, hl) )
+                return false;
+            return true;
+        }
+
+        if ( !g_tpm->hash(g_tpm, 2, base, size, &img_hl) )
+            return false;
+        for (unsigned int i=0; i<hl->count; i++) {
+            for (unsigned int j=0; j<img_hl.count; j++) {
+                if (hl->entries[i].alg == img_hl.entries[j].alg) {
+                    if ( !extend_hash(&hl->entries[i].hash,
+                            &img_hl.entries[j].hash, hl->entries[i].alg) )
+                        return false;
+                    break;
+                }
+            }
+        }
+
+        break;
+    }
+
+    case TB_EXTPOL_EMBEDDED: 
+    {
+        tb_hash_t img_hash;
+        hl->count = g_tpm->alg_count;
+        for (unsigned int i=0; i<hl->count; i++) {
+            hl->entries[i].alg = g_tpm->algs[i];
+            if ( !hash_buffer((const unsigned char *)cmdline, strlen(cmdline),
+                        &hl->entries[i].hash, g_tpm->algs[i]) )
+                return false;
+
+            if ( !hash_buffer(base, size, &img_hash, g_tpm->algs[i]) )
+                return false;
+            if ( !extend_hash(&hl->entries[i].hash, &img_hash, g_tpm->algs[i]) )
+                return false;
+        }
+
+        break;
+    }
+
+    default:
         return false;
+    }
 
     return true;
 }
 
 static bool is_hash_in_policy_entry(const tb_policy_entry_t *pol_entry,
-                                    tb_hash_t *hash, uint8_t hash_alg)
+                                    tb_hash_t *hash, uint16_t hash_alg)
 {
     /* assumes policy entry has been validated */
 
     if ( pol_entry == NULL || hash == NULL) {
-        printk("Error: input parameter is wrong.\n");
+        printk(TBOOT_ERR"Error: input parameter is wrong.\n");
         return false;
     }
 
@@ -423,9 +506,7 @@ static bool is_hash_in_policy_entry(const tb_policy_entry_t *pol_entry,
 }
 
 /*
- *
  * map policy type + error -> action
- *
  */
 static tb_policy_action_t evaluate_error(tb_error_t error)
 {
@@ -481,7 +562,7 @@ void apply_policy(tb_error_t error)
         case TB_POLACT_HALT:
             break; /* do halt at the end */
         default:
-            printk("Error: invalid policy action (%d)\n", action);
+            printk(TBOOT_ERR"Error: invalid policy action (%d)\n", action);
             /* do halt at the end */
     }
 
@@ -498,19 +579,34 @@ void apply_policy(tb_error_t error)
  * to PCR 18
  */
 static tb_error_t verify_module(module_t *module, tb_policy_entry_t *pol_entry,
-                                uint8_t hash_alg)
+                                uint16_t hash_alg)
 {
     /* assumes module is valid */
 
     void *base = (void *)module->mod_start;
     size_t size = module->mod_end - module->mod_start;
-    char *cmdline = (char *)module->string;
+    char *cmdline = get_module_cmd(g_ldr_ctx, module);
 
-    if ( pol_entry != NULL )
-        printk("verifying module \"%s\"...\n", cmdline);
-    tb_hash_t hash;
-    if ( !hash_module(&hash, TB_HALG_SHA1, cmdline, base, size) ) {
-        printk("\t hash cannot be generated.\n");
+    if ( pol_entry != NULL ) {
+        /* chunk the command line into 80 byte chunks */
+#define CHUNK_SIZE 80
+        int      cmdlen = strlen(cmdline);
+        char    *cptr = cmdline;
+        char     cmdchunk[CHUNK_SIZE+1];
+        printk(TBOOT_INFO"verifying module \"");
+        while (cmdlen > 0) {
+            strncpy(cmdchunk, cptr, CHUNK_SIZE);
+            cmdchunk[CHUNK_SIZE] = 0;
+            printk(TBOOT_INFO"\n%s", cmdchunk);
+            cmdlen -= CHUNK_SIZE;
+            cptr += CHUNK_SIZE;
+        }
+        printk(TBOOT_INFO"\"...\n");
+    }
+
+    hash_list_t hl;
+    if ( !hash_module(&hl, cmdline, base, size) ) {
+        printk(TBOOT_ERR"\t hash cannot be generated.\n");
         return TB_ERR_MODULE_VERIFICATION_FAILED;
     }
 
@@ -519,29 +615,34 @@ static tb_error_t verify_module(module_t *module, tb_policy_entry_t *pol_entry,
        PCRs won't match pre-S3
        NULL pol_entry means this is module 0 which is extended to PCR 18 */
     if ( NUM_VL_ENTRIES >= MAX_VL_HASHES )
-        printk("\t too many hashes to save\n");
+        printk(TBOOT_WARN"\t too many hashes to save\n");
     else if ( pol_entry == NULL || pol_entry->pcr != TB_POL_PCR_NONE ) {
         uint8_t pcr = (pol_entry == NULL ) ?
                           (g_using_da ? 17 : 18) : pol_entry->pcr;
         VL_ENTRIES(NUM_VL_ENTRIES).pcr = pcr;
-        VL_ENTRIES(NUM_VL_ENTRIES++).hash = hash;
+        VL_ENTRIES(NUM_VL_ENTRIES++).hl = hl;
     }
 
+    if ( g_tpm->extpol != TB_EXTPOL_FIXED )
+        return TB_ERR_NONE;
+
     if ( pol_entry != NULL &&
-         !is_hash_in_policy_entry(pol_entry, &hash, hash_alg) ) {
-        printk("\t verification failed\n");
+         !is_hash_in_policy_entry(pol_entry, &hl.entries[0].hash, hash_alg) ) {
+        printk(TBOOT_ERR"\t verification failed\n");
         return TB_ERR_MODULE_VERIFICATION_FAILED;
     }
 
     if ( pol_entry != NULL ) {
-        printk("\t OK : "); print_hash(&hash, TB_HALG_SHA1);
+        printk(TBOOT_DETA"\t OK : "); print_hash(&hl.entries[0].hash, hash_alg);
     }
+
     return TB_ERR_NONE;
 }
 
-void verify_all_modules(multiboot_info_t *mbi)
+static void verify_g_policy(void)
 {
     /* assumes mbi is valid */
+    printk(TBOOT_INFO"verifying policy \n");
 
     /* add entry for policy control field and (optionally) policy */
     /* hash will be <policy control field (4bytes)> | <hash policy> */
@@ -551,48 +652,213 @@ void verify_all_modules(multiboot_info_t *mbi)
     memcpy(buf, &g_policy->policy_control, sizeof(g_policy->policy_control));
     if ( g_policy->policy_control & TB_POLCTL_EXTEND_PCR17 ) {
         if ( !hash_policy((tb_hash_t *)&buf[sizeof(g_policy->policy_control)],
-                          TB_HALG_SHA1) ) {
-            printk("policy hash failed\n");
+                          g_tpm->cur_alg) ) {
+            printk(TBOOT_ERR"policy hash failed\n");
             apply_policy(TB_ERR_MODULE_VERIFICATION_FAILED);
         }
     }
-    if ( !hash_buffer(buf, get_hash_size(TB_HALG_SHA1) +
-                           sizeof(g_policy->policy_control),
-                      &VL_ENTRIES(NUM_VL_ENTRIES).hash, TB_HALG_SHA1) )
+
+    u32 size = get_hash_size(g_tpm->cur_alg) + sizeof(g_policy->policy_control);
+    switch (g_tpm->extpol) {
+    case TB_EXTPOL_FIXED: 
+        VL_ENTRIES(NUM_VL_ENTRIES).hl.count = 1;
+        VL_ENTRIES(NUM_VL_ENTRIES).hl.entries[0].alg = g_tpm->cur_alg;
+        if ( !hash_buffer(buf, size,
+                &VL_ENTRIES(NUM_VL_ENTRIES).hl.entries[0].hash, g_tpm->cur_alg) )
+            apply_policy(TB_ERR_MODULE_VERIFICATION_FAILED);
+
+        break;
+
+    case TB_EXTPOL_AGILE: 
+        if ( !g_tpm->hash(g_tpm, 2, buf, size, &VL_ENTRIES(NUM_VL_ENTRIES).hl) )
+            apply_policy(TB_ERR_MODULE_VERIFICATION_FAILED);
+        break;
+
+    case TB_EXTPOL_EMBEDDED: 
+    {
+        VL_ENTRIES(NUM_VL_ENTRIES).hl.count = g_tpm->alg_count;
+        for (int i=0; i<g_tpm->alg_count; i++) {
+            VL_ENTRIES(NUM_VL_ENTRIES).hl.entries[i].alg = g_tpm->algs[i];
+            if ( !hash_buffer(buf, size, &VL_ENTRIES(NUM_VL_ENTRIES).hl.entries[i].hash,
+                        g_tpm->algs[i]) )
+                return;
+        }
+
+        break;
+    }
+
+    default:
         apply_policy(TB_ERR_MODULE_VERIFICATION_FAILED);
+        break;
+    }
+
     VL_ENTRIES(NUM_VL_ENTRIES++).pcr = 17;
     if ( g_using_da ) {
         /* copying hash of policy_control into PCR 18 */
         if ( NUM_VL_ENTRIES >= MAX_VL_HASHES )
-            printk("\t too many hashes to save for DA\n");
+            printk(TBOOT_ERR"\t too many hashes to save for DA\n");
         else {
-            VL_ENTRIES(NUM_VL_ENTRIES).hash = VL_ENTRIES(NUM_VL_ENTRIES-1).hash;
+            VL_ENTRIES(NUM_VL_ENTRIES).hl = VL_ENTRIES(NUM_VL_ENTRIES-1).hl;
             VL_ENTRIES(NUM_VL_ENTRIES++).pcr = 18;
         }
     }
+} 
+
+void verify_all_modules(loader_ctx *lctx)
+{
+    /* assumes mbi is valid */
+    verify_g_policy();
 
     /* module 0 is always extended to PCR 18, so add entry for it */
-    apply_policy(verify_module(get_module(mbi, 0), NULL, g_policy->hash_alg));
+    apply_policy(verify_module(get_module(lctx, 0), NULL, g_policy->hash_alg));
 
     /* now verify each module and add its hash */
-    for ( unsigned int i = 0; i < mbi->mods_count; i++ ) {
-        module_t *module = get_module(mbi, i);
+    for ( unsigned int i = 0; i < get_module_count(lctx); i++ ) {
+        module_t *module = get_module(lctx, i);
         tb_policy_entry_t *pol_entry = find_policy_entry(g_policy, i);
         if ( module == NULL ) {
-            printk("missing module entry %u\n", i);
+            printk(TBOOT_ERR"missing module entry %u\n", i);
             apply_policy(TB_ERR_MODULE_VERIFICATION_FAILED);
         }
         else if ( pol_entry == NULL ) {
-            printk("policy entry for module %u not found\n", i);
+            printk(TBOOT_ERR"policy entry for module %u not found\n", i);
             apply_policy(TB_ERR_MODULES_NOT_IN_POLICY);
         }
         else
             apply_policy(verify_module(module, pol_entry, g_policy->hash_alg));
     }
 
-    printk("all modules are verified\n");
+    printk(TBOOT_INFO"all modules are verified\n");
 }
 
+static int find_first_nvpolicy_entry(const tb_policy_t *policy)
+{
+    if ( policy == NULL ) {
+        PRINT(TBOOT_ERR"Error: policy pointer is NULL\n");
+        return -1;
+    }
+
+    for ( int i = 0; i < policy->num_entries; i++ ) {
+        tb_policy_entry_t *pol_entry = get_policy_entry(policy, i);
+        if ( pol_entry == NULL )
+            return -1;
+
+        if ( pol_entry->mod_num == TB_POL_MOD_NUM_NV ||
+             pol_entry->mod_num == TB_POL_MOD_NUM_NV_RAW )
+            return i;
+    }
+
+    return -1;
+}
+
+static int find_next_nvpolicy_entry(const tb_policy_t *policy, int i)
+{
+    if ( policy == NULL || i < 0 || i >= policy->num_entries )
+        return -1;
+
+    for ( i++; i < policy->num_entries; i++ ) {
+        tb_policy_entry_t *pol_entry = get_policy_entry(policy, i);
+        if ( pol_entry == NULL )
+            return -1;
+
+        if ( pol_entry->mod_num == TB_POL_MOD_NUM_NV ||
+             pol_entry->mod_num == TB_POL_MOD_NUM_NV_RAW )
+            return i;
+    }
+
+    return -1;
+}
+
+static uint8_t nv_buf[4096];
+
+static tb_error_t verify_nvindex(tb_policy_entry_t *pol_entry,
+                                 uint16_t hash_alg)
+{
+    size_t nv_size = sizeof(nv_buf);
+    tb_hash_t digest;
+    uint32_t attribute;
+
+    if ( pol_entry == NULL )
+        return TB_ERR_NV_VERIFICATION_FAILED;
+
+    printk(TBOOT_INFO"verifying nv index 0x%08X\n", pol_entry->nv_index);
+
+    /* check nv attribute */
+    if ( !g_tpm->get_nvindex_permission(g_tpm, 0, pol_entry->nv_index,
+                                                   &attribute) ) {
+        printk(TBOOT_ERR"\t :reading nv index permission failed\n");
+        return TB_ERR_NV_VERIFICATION_FAILED;
+    }
+    if ( !(attribute & TPM_NV_PER_OWNERWRITE) ) {
+        printk(TBOOT_ERR"\t :nv index should be OWNERWRITE, bad permission\n");
+        return TB_ERR_NV_VERIFICATION_FAILED;
+    }
+
+    /* get nv content */
+    memset(nv_buf, 0, sizeof(nv_buf));
+    if ( !read_policy_from_tpm(pol_entry->nv_index,
+                nv_buf, &nv_size) ) {
+        printk(TBOOT_ERR"\t :reading nv index 0x%08X failed\n",
+               pol_entry->nv_index);
+        return TB_ERR_NV_VERIFICATION_FAILED;
+    }
+
+    /* hash the buffer if needed */
+    switch ( pol_entry->mod_num ) {
+    case TB_POL_MOD_NUM_NV:
+        if ( !hash_buffer((const uint8_t*)nv_buf, nv_size, &digest,
+                          TB_HALG_SHA1) ) {
+            printk(TBOOT_ERR"\t :nv content hash failed\n");
+            return TB_ERR_NV_VERIFICATION_FAILED;
+        }
+        break;
+    case TB_POL_MOD_NUM_NV_RAW:
+        if ( nv_size != sizeof(digest.sha1) ) {
+            printk(TBOOT_ERR"\t :raw nv with wrong size (%d), should be %d\n",
+                   (int)nv_size, sizeof(digest.sha1));
+            return TB_ERR_NV_VERIFICATION_FAILED;
+        }
+        memcpy(digest.sha1, nv_buf, nv_size);
+        break;
+    default:
+        printk(TBOOT_ERR"\t :bad mod number for NV measuring in policy entry: %d\n",
+               pol_entry->mod_num);
+        return TB_ERR_NV_VERIFICATION_FAILED;
+    }
+
+    /* add new hash to list (unless it doesn't get put in a PCR)
+       we'll just drop it if the list is full, but that will mean S3 resume
+       PCRs won't match pre-S3 */
+    if ( NUM_VL_ENTRIES >= MAX_VL_HASHES )
+        printk(TBOOT_WARN"\t too many hashes to save\n");
+    else if ( pol_entry->pcr != TB_POL_PCR_NONE ) {
+        VL_ENTRIES(NUM_VL_ENTRIES).pcr = pol_entry->pcr;
+        VL_ENTRIES(NUM_VL_ENTRIES).hl.count = 1;
+        VL_ENTRIES(NUM_VL_ENTRIES).hl.entries[0].alg = TB_HALG_SHA1;
+        memcpy(VL_ENTRIES(NUM_VL_ENTRIES++).hl.entries[0].hash.sha1,
+                digest.sha1, SHA1_LENGTH);
+    }
+
+    /* verify nv content */
+    if ( !is_hash_in_policy_entry(pol_entry, &digest, hash_alg) ) {
+        printk(TBOOT_ERR"\t verification failed\n");
+        return TB_ERR_NV_VERIFICATION_FAILED;
+    }
+
+    printk(TBOOT_DETA"\t OK : "); print_hash(&digest, hash_alg);
+    return TB_ERR_NONE;
+}
+
+void verify_all_nvindices(void)
+{
+    /* go through nv policies in tb policy */
+    for ( int i = find_first_nvpolicy_entry(g_policy);
+          i >= 0;
+          i = find_next_nvpolicy_entry(g_policy, i) ) {
+        tb_policy_entry_t *pol_entry = get_policy_entry(g_policy, i);
+        apply_policy(verify_nvindex(pol_entry, g_policy->hash_alg));
+    }
+}
 
 /*
  * Local variables:
