@@ -38,7 +38,6 @@
 #include <stdbool.h>
 #include <types.h>
 #include <tb_error.h>
-#include <multiboot.h>
 #include <msr.h>
 #include <compiler.h>
 #include <string.h>
@@ -49,14 +48,15 @@
 #include <atomic.h>
 #include <mutex.h>
 #include <tpm.h>
-#include <e820.h>
 #include <uuid.h>
 #include <loader.h>
+#include <e820.h>
 #include <tboot.h>
 #include <mle.h>
 #include <hash.h>
 #include <lcp2.h>
 #include <cmdline.h>
+#include <acpi.h>
 #include <txt/txt.h>
 #include <txt/config_regs.h>
 #include <txt/mtrrs.h>
@@ -65,6 +65,7 @@
 #include <txt/smx.h>
 #include <txt/verify.h>
 #include <txt/vmcs.h>
+#include <io.h>
 
 /* counter timeout for waiting for all APs to enter wait-for-sipi */
 #define AP_WFS_TIMEOUT     0x01000000
@@ -86,6 +87,7 @@ extern tboot_shared_t _tboot_shared;
 extern void apply_policy(tb_error_t error);
 extern void cpu_wakeup(uint32_t cpuid, uint32_t sipi_vec);
 extern void print_event(const tpm12_pcr_event_t *evt);
+extern void print_event_2(void *evt, uint16_t alg);
 
 /*
  * this is the structure whose addr we'll put in TXT heap
@@ -113,26 +115,27 @@ atomic_t ap_wfs_count;
 
 static void print_file_info(void)
 {
-    printk("file addresses:\n");
-    printk("\t &_start=%p\n", &_start);
-    printk("\t &_end=%p\n", &_end);
-    printk("\t &_mle_start=%p\n", &_mle_start);
-    printk("\t &_mle_end=%p\n", &_mle_end);
-    printk("\t &_post_launch_entry=%p\n", &_post_launch_entry);
-    printk("\t &_txt_wakeup=%p\n", &_txt_wakeup);
-    printk("\t &g_mle_hdr=%p\n", &g_mle_hdr);
+    printk(TBOOT_DETA"file addresses:\n");
+    printk(TBOOT_DETA"\t &_start=%p\n", &_start);
+    printk(TBOOT_DETA"\t &_end=%p\n", &_end);
+    printk(TBOOT_DETA"\t &_mle_start=%p\n", &_mle_start);
+    printk(TBOOT_DETA"\t &_mle_end=%p\n", &_mle_end);
+    printk(TBOOT_DETA"\t &_post_launch_entry=%p\n", &_post_launch_entry);
+    printk(TBOOT_DETA"\t &_txt_wakeup=%p\n", &_txt_wakeup);
+    printk(TBOOT_DETA"\t &g_mle_hdr=%p\n", &g_mle_hdr);
 }
 
 static void print_mle_hdr(const mle_hdr_t *mle_hdr)
 {
-    printk("MLE header:\n");
-    printk("\t uuid="); print_uuid(&mle_hdr->uuid); printk("\n");
-    printk("\t length=%x\n", mle_hdr->length);
-    printk("\t version=%08x\n", mle_hdr->version);
-    printk("\t entry_point=%08x\n", mle_hdr->entry_point);
-    printk("\t first_valid_page=%08x\n", mle_hdr->first_valid_page);
-    printk("\t mle_start_off=%x\n", mle_hdr->mle_start_off);
-    printk("\t mle_end_off=%x\n", mle_hdr->mle_end_off);
+    printk(TBOOT_DETA"MLE header:\n");
+    printk(TBOOT_DETA"\t uuid="); print_uuid(&mle_hdr->uuid); 
+    printk(TBOOT_DETA"\n");
+    printk(TBOOT_DETA"\t length=%x\n", mle_hdr->length);
+    printk(TBOOT_DETA"\t version=%08x\n", mle_hdr->version);
+    printk(TBOOT_DETA"\t entry_point=%08x\n", mle_hdr->entry_point);
+    printk(TBOOT_DETA"\t first_valid_page=%08x\n", mle_hdr->first_valid_page);
+    printk(TBOOT_DETA"\t mle_start_off=%x\n", mle_hdr->mle_start_off);
+    printk(TBOOT_DETA"\t mle_end_off=%x\n", mle_hdr->mle_end_off);
     print_txt_caps("\t ", mle_hdr->capabilities);
 }
 
@@ -148,7 +151,8 @@ static void print_mle_hdr(const mle_hdr_t *mle_hdr)
 /* 1 ptable = 3 pages and just 1 loop loop for ptable MLE page table */
 /* can only contain 4k pages */
 
-static __mlept uint8_t g_mle_pt[3 * PAGE_SIZE];  /* pgdir ptr + pgdir + ptab = 3 */
+static __mlept uint8_t g_mle_pt[3 * PAGE_SIZE];  
+/* pgdir ptr + pgdir + ptab = 3 */
 
 static void *build_mle_pagetable(uint32_t mle_start, uint32_t mle_size)
 {
@@ -157,17 +161,17 @@ static void *build_mle_pagetable(uint32_t mle_start, uint32_t mle_size)
     void *pg_dir_ptr_tab, *pg_dir, *pg_tab;
     uint64_t *pte;
 
-    printk("MLE start=%x, end=%x, size=%x\n", mle_start, mle_start+mle_size,
-           mle_size);
+    printk(TBOOT_DETA"MLE start=%x, end=%x, size=%x\n", 
+           mle_start, mle_start+mle_size, mle_size);
     if ( mle_size > 512*PAGE_SIZE ) {
-        printk("MLE size too big for single page table\n");
+        printk(TBOOT_ERR"MLE size too big for single page table\n");
         return NULL;
     }
 
 
     /* should start on page boundary */
     if ( mle_start & ~PAGE_MASK ) {
-        printk("MLE start is not page-aligned\n");
+        printk(TBOOT_ERR"MLE start is not page-aligned\n");
         return NULL;
     }
 
@@ -175,7 +179,7 @@ static void *build_mle_pagetable(uint32_t mle_start, uint32_t mle_size)
     ptab_size = sizeof(g_mle_pt);
     ptab_base = &g_mle_pt;
     memset(ptab_base, 0, ptab_size);
-    printk("ptab_size=%x, ptab_base=%p\n", ptab_size, ptab_base);
+    printk(TBOOT_DETA"ptab_size=%x, ptab_base=%p\n", ptab_size, ptab_base);
 
     pg_dir_ptr_tab = ptab_base;
     pg_dir         = pg_dir_ptr_tab + PAGE_SIZE;
@@ -199,83 +203,9 @@ static void *build_mle_pagetable(uint32_t mle_start, uint32_t mle_size)
     return ptab_base;
 }
 
-/*
- * will go through all modules to find an SINIT that matches the platform
- * (size can be NULL)
- */
-static bool find_platform_sinit_module(const multiboot_info_t *mbi, void **base,
-                                       uint32_t *size)
-{
-    if ( base != NULL )
-        *base = NULL;
-    if ( size != NULL )
-        *size = 0;
-
-    if ( mbi->mods_addr == 0 || mbi->mods_count == 0 ) {
-        printk("no module info\n");
-        return false;
-    }
-
-    for ( unsigned int i = mbi->mods_count - 1; i > 0; i-- ) {
-        module_t *m = get_module(mbi, i);
-
-        printk("checking if module %s is an SINIT for this platform...\n",
-               (const char *)m->string);
-        void *base2 = (void *)m->mod_start;
-        uint32_t size2 = m->mod_end - (unsigned long)(base2);
-        if ( is_sinit_acmod(base2, size2, false) &&
-             does_acmod_match_platform((acm_hdr_t *)base2) ) {
-            if ( base != NULL )
-                *base = base2;
-            if ( size != NULL )
-                *size = size2;
-            printk("SINIT matches platform\n");
-            return true;
-        }
-    }
-    /* no SINIT found for this platform */
-    printk("no SINIT AC module found\n");
-    return false;
-}
-
-bool find_lcp_module(const multiboot_info_t *mbi, void **base, uint32_t *size)
-{
-    size_t size2 = 0;
-    void *base2 = NULL;
-
-    if ( base != NULL )
-        *base = NULL;
-    if ( size != NULL )
-        *size = 0;
-
-    /* try policy data file for old version (0x00 or 0x01) */
-    find_module_by_uuid(mbi, &base2, &size2, &((uuid_t)LCP_POLICY_DATA_UUID));
-
-    /* not found */
-    if ( base2 == NULL ) {
-        /* try policy data file for new version (0x0202) */
-        find_module_by_file_signature(mbi, &base2, &size2,
-                                      LCP_POLICY_DATA_FILE_SIGNATURE);
-
-        if ( base2 == NULL ) {
-            printk("no LCP module found\n");
-            return false;
-        }
-        else
-            printk("v2 LCP policy data found\n");
-    }
-    else
-        printk("v1 LCP policy data found\n");
-
-
-    if ( base != NULL )
-        *base = base2;
-    if ( size != NULL )
-        *size = size2;
-    return true;
-}
 
 static __data event_log_container_t *g_elog = NULL;
+static __data heap_event_log_ptr_elt2_t *g_elog_2 = NULL;
 
 /* should be called after os_mle_data initialized */
 static void *init_event_log(void)
@@ -296,6 +226,63 @@ static void *init_event_log(void)
     return (void *)g_elog;
 }
 
+static void init_evtlog_desc(heap_event_log_ptr_elt2_t *evt_log)
+{
+    unsigned int i;
+    os_mle_data_t *os_mle_data = get_os_mle_data_start(get_txt_heap());
+
+    switch (g_tpm->extpol) {
+    case TB_EXTPOL_AGILE:
+        for (i=0; i<evt_log->count; i++) {
+            evt_log->event_log_descr[i].alg = g_tpm->algs_banks[i];
+            evt_log->event_log_descr[i].phys_addr =
+                    (uint64_t)(unsigned long)(os_mle_data->event_log_buffer + i*4096);
+            evt_log->event_log_descr[i].size = 4096;
+            evt_log->event_log_descr[i].pcr_events_offset = 0;
+            evt_log->event_log_descr[i].next_event_offset = 0;
+            if (g_tpm->algs_banks[i] != TB_HALG_SHA1) {
+                evt_log->event_log_descr[i].pcr_events_offset =
+                        32 + sizeof(tpm20_log_descr_t);
+                evt_log->event_log_descr[i].next_event_offset =
+                        32 + sizeof(tpm20_log_descr_t);
+            }
+        }
+        break;
+    case TB_EXTPOL_EMBEDDED:
+        for (i=0; i<evt_log->count; i++) {
+            evt_log->event_log_descr[i].alg = g_tpm->algs[i];
+            evt_log->event_log_descr[i].phys_addr =
+                    (uint64_t)(unsigned long)(os_mle_data->event_log_buffer + i*4096);
+            evt_log->event_log_descr[i].size = 4096;
+            evt_log->event_log_descr[i].pcr_events_offset = 0;
+            evt_log->event_log_descr[i].next_event_offset = 0;
+            if (g_tpm->algs[i] != TB_HALG_SHA1) {
+                evt_log->event_log_descr[i].pcr_events_offset =
+                        32 + sizeof(tpm20_log_descr_t);
+                evt_log->event_log_descr[i].next_event_offset =
+                        32 + sizeof(tpm20_log_descr_t);
+            }
+        }
+        break;
+    case TB_EXTPOL_FIXED:
+        evt_log->event_log_descr[0].alg = g_tpm->cur_alg;
+        evt_log->event_log_descr[0].phys_addr =
+                    (uint64_t)(unsigned long)os_mle_data->event_log_buffer;
+        evt_log->event_log_descr[0].size = 4096;
+        evt_log->event_log_descr[0].pcr_events_offset = 0;
+        evt_log->event_log_descr[0].next_event_offset = 0;
+        if (g_tpm->cur_alg != TB_HALG_SHA1) {
+            evt_log->event_log_descr[0].pcr_events_offset =
+                    32 + sizeof(tpm20_log_descr_t);
+            evt_log->event_log_descr[0].next_event_offset =
+                    32 + sizeof(tpm20_log_descr_t);
+        }
+        break;
+    default:
+        return;
+    }
+}
+
 static void init_os_sinit_ext_data(heap_ext_data_element_t* elts)
 {
     heap_ext_data_element_t* elt = elts;
@@ -306,12 +293,29 @@ static void init_os_sinit_ext_data(heap_ext_data_element_t* elts)
     elt->type = HEAP_EXTDATA_TYPE_TPM_EVENT_LOG_PTR;
     elt->size = sizeof(*elt) + sizeof(*evt_log);
 
+    if ( g_tpm->major == TPM20_VER_MAJOR ) {
+        g_elog_2 = (heap_event_log_ptr_elt2_t *)elt->data;
+
+        if ( g_tpm->extpol == TB_EXTPOL_AGILE )
+            g_elog_2->count = g_tpm->banks;
+        else if ( g_tpm->extpol == TB_EXTPOL_EMBEDDED )
+            g_elog_2->count = g_tpm->alg_count;
+        else
+            g_elog_2->count = 1;
+
+        init_evtlog_desc(g_elog_2);
+
+        elt->type = HEAP_EXTDATA_TYPE_TPM_EVENT_LOG_PTR_2;
+        elt->size = sizeof(*elt) + sizeof(u32) +
+                g_elog_2->count * sizeof(heap_event_log_descr_t);
+    }
+
     elt = (void *)elt + elt->size;
     elt->type = HEAP_EXTDATA_TYPE_END;
     elt->size = sizeof(*elt);
 }
 
-bool evtlog_append(uint8_t pcr, tb_hash_t *hash, uint32_t type)
+bool evtlog_append_tpm12(uint8_t pcr, tb_hash_t *hash, uint32_t type)
 {
     if ( g_elog == NULL )
         return true;
@@ -333,13 +337,99 @@ bool evtlog_append(uint8_t pcr, tb_hash_t *hash, uint32_t type)
     return true;
 }
 
+void dump_event_2(void)
+{
+    heap_event_log_descr_t *log_descr;
+
+    for ( unsigned int i=0; i<g_elog_2->count; i++ ) {
+        log_descr = &g_elog_2->event_log_descr[i];
+        printk(TBOOT_DETA"\t\t\t Log Descrption:\n");
+        printk(TBOOT_DETA"\t\t\t             Alg: %u\n", log_descr->alg);
+        printk(TBOOT_DETA"\t\t\t            Size: %u\n", log_descr->size);
+        printk(TBOOT_DETA"\t\t\t    EventsOffset: [%u,%u)\n",
+                log_descr->pcr_events_offset,
+                log_descr->next_event_offset);
+
+        uint32_t hash_size, data_size; 
+        hash_size = get_hash_size(log_descr->alg); 
+        if ( hash_size == 0 )
+            return;
+
+        void *curr, *next;
+        *((u64 *)(&curr)) = log_descr->phys_addr +
+                log_descr->pcr_events_offset;
+        *((u64 *)(&next)) = log_descr->phys_addr +
+                log_descr->next_event_offset;
+
+        while ( curr < next ) {
+            print_event_2(curr, log_descr->alg);
+            data_size = *(uint32_t *)(curr + 2*sizeof(uint32_t) + hash_size);
+            curr += 3*sizeof(uint32_t) + hash_size + data_size;
+        }
+    }
+}
+
+bool evtlog_append_tpm20(uint8_t pcr, uint16_t alg, tb_hash_t *hash, uint32_t type)
+{
+    heap_event_log_descr_t *cur_desc = NULL;
+    uint32_t hash_size; 
+    void *cur, *next;
+
+    for ( unsigned int i=0; i<g_elog_2->count; i++ ) {
+        if ( g_elog_2->event_log_descr[i].alg == alg ) {
+            cur_desc = &g_elog_2->event_log_descr[i];
+            break;
+        }
+    }
+    if ( !cur_desc )
+        return false;
+   
+    hash_size = get_hash_size(alg); 
+    if ( hash_size == 0 )
+        return false;
+    *((u64 *)(&cur)) = *((u64 *)(&next)) =
+          cur_desc->phys_addr + cur_desc->next_event_offset;
+    if ( cur_desc->next_event_offset + 32 > cur_desc->size )
+        return false;
+    *((u32 *)next) = pcr;
+    next += sizeof(u32);
+    *((u32 *)next) = type;
+    next += sizeof(u32);
+    memcpy((uint8_t *)next, hash, hash_size);
+    next += hash_size/sizeof(uint32_t);
+    *((u32 *)next) = 0;
+    cur_desc->next_event_offset += 3*sizeof(uint32_t) + hash_size; 
+
+    print_event_2(cur, alg);
+    return true;
+}
+
+bool evtlog_append(uint8_t pcr, hash_list_t *hl, uint32_t type)
+{
+    switch (g_tpm->major) {
+    case TPM12_VER_MAJOR:
+        evtlog_append_tpm12(pcr, &hl->entries[0].hash, type);
+        break;
+    case TPM20_VER_MAJOR:
+        for (unsigned int i=0; i<hl->count; i++)
+            evtlog_append_tpm20(pcr, hl->entries[i].alg,
+                    &hl->entries[i].hash, type);
+        break;
+    default:
+        return false;
+    }
+
+    return true;
+}
+
 __data uint32_t g_using_da = 0;
+__data acm_hdr_t *g_sinit = 0;
 
 /*
  * sets up TXT heap
  */
 static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
-                                 const multiboot_info_t *mbi)
+                                 loader_ctx *lctx)
 {
     txt_heap_t *txt_heap;
     uint64_t *size;
@@ -360,7 +450,7 @@ static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
     *size = sizeof(*os_mle_data) + sizeof(uint64_t);
     memset(os_mle_data, 0, sizeof(*os_mle_data));
     os_mle_data->version = 3;
-    os_mle_data->mbi = (multiboot_info_t *)(unsigned long)mbi;
+    os_mle_data->lctx_addr = lctx->addr;
     os_mle_data->saved_misc_enable_msr = rdmsr(MSR_IA32_MISC_ENABLE);
 
     /*
@@ -369,7 +459,8 @@ static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
     /* check sinit supported os_sinit_data version */
     uint32_t version = get_supported_os_sinit_data_ver(sinit);
     if ( version < MIN_OS_SINIT_DATA_VER ) {
-        printk("unsupported OS to SINIT data version(%u) in sinit\n", version);
+        printk(TBOOT_ERR"unsupported OS to SINIT data version(%u) in sinit\n",
+               version);
         return NULL;
     }
     if ( version > MAX_OS_SINIT_DATA_VER )
@@ -389,17 +480,21 @@ static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
         (uint64_t)(unsigned long)&_mle_start;
     /* VT-d PMRs */
     uint64_t min_lo_ram, max_lo_ram, min_hi_ram, max_hi_ram;
+    
     if ( !get_ram_ranges(&min_lo_ram, &max_lo_ram, &min_hi_ram, &max_hi_ram) )
         return NULL;
+
     set_vtd_pmrs(os_sinit_data, min_lo_ram, max_lo_ram, min_hi_ram,
                  max_hi_ram);
     /* LCP owner policy data */
     void *lcp_base = NULL;
     uint32_t lcp_size = 0;
-    if ( find_lcp_module(mbi, &lcp_base, &lcp_size) && lcp_size > 0 ) {
+
+    if ( find_lcp_module(lctx, &lcp_base, &lcp_size) && lcp_size > 0 ) {
         /* copy to heap */
         if ( lcp_size > sizeof(os_mle_data->lcp_po_data) ) {
-            printk("LCP owner policy data file is too large (%u)\n", lcp_size);
+            printk(TBOOT_ERR"LCP owner policy data file is too large (%u)\n",
+                   lcp_size);
             return NULL;
         }
         memcpy(os_mle_data->lcp_po_data, lcp_base, lcp_size);
@@ -418,7 +513,8 @@ static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
     else if ( sinit_caps.rlp_wake_getsec )
         os_sinit_data->capabilities.rlp_wake_getsec = 1;
     else {     /* should have been detected in verify_acmod() */
-        printk("SINIT capabilities are icompatible (0x%x)\n", sinit_caps._raw);
+        printk(TBOOT_ERR"SINIT capabilities are incompatible (0x%x)\n", 
+               sinit_caps._raw);
         return NULL;
     }
     /* capabilities : require MLE pagetable in ECX on launch */
@@ -426,7 +522,25 @@ static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
      * os_sinit_data->capabilities.ecx_pgtbl = 1;
      */
     os_sinit_data->capabilities.ecx_pgtbl = 0;
-    /* TODO: when tboot supports EFI then set efi_rsdt_ptr */
+    if (is_loader_launch_efi(lctx)){
+        /* we were launched EFI, set efi_rsdt_ptr */
+        struct acpi_rsdp *rsdp = get_rsdp(lctx);
+        if (rsdp != NULL){
+            if (version < 6){
+                /* rsdt */
+                /* NOTE: Winston Wang says this doesn't work for v5 */
+                os_sinit_data->efi_rsdt_ptr = (uint64_t) rsdp->rsdp1.rsdt;
+            } else {
+                /* rsdp */
+                os_sinit_data->efi_rsdt_ptr = (uint64_t)((uint32_t) rsdp);
+            }
+        } else {
+            /* per discussions--if we don't have an ACPI pointer, die */
+            printk(TBOOT_ERR"Failed to find RSDP for EFI launch\n");
+            return NULL;
+        }
+    }
+        
     /* capabilities : choose DA/LG */
     os_sinit_data->capabilities.pcr_map_no_legacy = 1;
     if ( sinit_caps.pcr_map_da && get_tboot_prefer_da() )
@@ -434,15 +548,25 @@ static txt_heap_t *init_txt_heap(void *ptab_base, acm_hdr_t *sinit,
     else if ( !sinit_caps.pcr_map_no_legacy )
         os_sinit_data->capabilities.pcr_map_no_legacy = 0;
     else if ( sinit_caps.pcr_map_da ) {
-        printk("DA is the only supported PCR mapping by SINIT, use it\n");
+        printk(TBOOT_INFO
+               "DA is the only supported PCR mapping by SINIT, use it\n");
         os_sinit_data->capabilities.pcr_map_da = 1;
     }
     else {
-        printk("SINIT capabilities are icompatible (0x%x)\n", sinit_caps._raw);
+        printk(TBOOT_ERR"SINIT capabilities are incompatible (0x%x)\n", 
+               sinit_caps._raw);
         return NULL;
     }
     g_using_da = os_sinit_data->capabilities.pcr_map_da;
-        
+
+    /* PCR mapping selection MUST be zero in TPM2.0 mode
+     * since D/A mapping is the only supported by TPM2.0 */
+    if ( g_tpm->major >= TPM20_VER_MAJOR ) {
+        os_sinit_data->flags = (g_tpm->extpol == TB_EXTPOL_AGILE) ? 0 : 1;
+        os_sinit_data->capabilities.pcr_map_no_legacy = 0;
+        os_sinit_data->capabilities.pcr_map_da = 0;
+        g_using_da = 1;
+    }   
 
     /* Event log initialization */
     if ( os_sinit_data->version >= 6 )
@@ -468,7 +592,7 @@ static void txt_wakeup_cpus(void)
 
     /* enable SMIs on BSP before waking APs (which will enable them on APs)
        because some SMM may take immediate SMI and hang if AP gets in first */
-    printk("enabling SMIs on BSP\n");
+    printk(TBOOT_DETA"enabling SMIs on BSP\n");
     __getsec_smctrl();
 
     atomic_set(&ap_wfs_count, 0);
@@ -482,10 +606,10 @@ static void txt_wakeup_cpus(void)
     mle_join.gdt_base = (uint32_t)gdt_table;
     mle_join.gdt_limit = gdt_table_end - gdt_table - 1;
 
-    printk("mle_join.entry_point = %x\n", mle_join.entry_point);
-    printk("mle_join.seg_sel = %x\n", mle_join.seg_sel);
-    printk("mle_join.gdt_base = %x\n", mle_join.gdt_base);
-    printk("mle_join.gdt_limit = %x\n", mle_join.gdt_limit);
+    printk(TBOOT_DETA"mle_join.entry_point = %x\n", mle_join.entry_point);
+    printk(TBOOT_DETA"mle_join.seg_sel = %x\n", mle_join.seg_sel);
+    printk(TBOOT_DETA"mle_join.gdt_base = %x\n", mle_join.gdt_base);
+    printk(TBOOT_DETA"mle_join.gdt_limit = %x\n", mle_join.gdt_limit);
 
     write_priv_config_reg(TXTCR_MLE_JOIN, (uint64_t)(unsigned long)&mle_join);
 
@@ -497,14 +621,15 @@ static void txt_wakeup_cpus(void)
 
     /* choose wakeup mechanism based on capabilities used */
     if ( os_sinit_data->capabilities.rlp_wake_monitor ) {
-        printk("joining RLPs to MLE with MONITOR wakeup\n");
-        printk("rlp_wakeup_addr = 0x%x\n", sinit_mle_data->rlp_wakeup_addr);
+        printk(TBOOT_INFO"joining RLPs to MLE with MONITOR wakeup\n");
+        printk(TBOOT_DETA"rlp_wakeup_addr = 0x%x\n", 
+               sinit_mle_data->rlp_wakeup_addr);
         *((uint32_t *)(unsigned long)(sinit_mle_data->rlp_wakeup_addr)) = 0x01;
     }
     else {
-        printk("joining RLPs to MLE with GETSEC[WAKEUP]\n");
+        printk(TBOOT_INFO"joining RLPs to MLE with GETSEC[WAKEUP]\n");
         __getsec_wakeup();
-        printk("GETSEC[WAKEUP] completed\n");
+        printk(TBOOT_INFO"GETSEC[WAKEUP] completed\n");
     }
 
     /* assume BIOS isn't lying to us about # CPUs, else some CPUS may not */
@@ -514,29 +639,29 @@ static void txt_wakeup_cpus(void)
     bios_data_t *bios_data = get_bios_data_start(txt_heap);
     ap_wakeup_count = bios_data->num_logical_procs - 1;
     if ( ap_wakeup_count >= NR_CPUS ) {
-        printk("there are too many CPUs (%u)\n", ap_wakeup_count);
+        printk(TBOOT_INFO"there are too many CPUs (%u)\n", ap_wakeup_count);
         ap_wakeup_count = NR_CPUS - 1;
     }
 
-    printk("waiting for all APs (%d) to enter wait-for-sipi...\n",
+    printk(TBOOT_INFO"waiting for all APs (%d) to enter wait-for-sipi...\n",
            ap_wakeup_count);
     /* wait for all APs that woke up to have entered wait-for-sipi */
     uint32_t timeout = AP_WFS_TIMEOUT;
     do {
         if ( timeout % 0x8000 == 0 )
-            printk(".");
+            printk(TBOOT_INFO".");
         else
             cpu_relax();
         if ( timeout % 0x200000 == 0 )
-            printk("\n");
+            printk(TBOOT_INFO"\n");
         timeout--;
     } while ( ( atomic_read(&ap_wfs_count) < ap_wakeup_count ) &&
               timeout > 0 );
-    printk("\n");
+    printk(TBOOT_INFO"\n");
     if ( timeout == 0 )
-        printk("wait-for-sipi loop timed-out\n");
+        printk(TBOOT_INFO"wait-for-sipi loop timed-out\n");
     else
-        printk("all APs in wait-for-sipi\n");
+        printk(TBOOT_INFO"all APs in wait-for-sipi\n");
 }
 
 bool txt_is_launched(void)
@@ -548,9 +673,8 @@ bool txt_is_launched(void)
     return sts.senter_done_sts;
 }
 
-tb_error_t txt_launch_environment(const multiboot_info_t *mbi)
+tb_error_t txt_launch_environment(loader_ctx *lctx)
 {
-    acm_hdr_t *sinit = NULL;
     void *mle_ptab_base;
     os_mle_data_t *os_mle_data;
     txt_heap_t *txt_heap;
@@ -558,14 +682,14 @@ tb_error_t txt_launch_environment(const multiboot_info_t *mbi)
     /*
      * find correct SINIT AC module in modules list
      */
-    find_platform_sinit_module(mbi, (void **)&sinit, NULL);
+    find_platform_sinit_module(lctx, (void **)&g_sinit, NULL);
     /* if it is newer than BIOS-provided version, then copy it to */
     /* BIOS reserved region */
-    sinit = copy_sinit(sinit);
-    if ( sinit == NULL )
+    g_sinit = copy_sinit(g_sinit);
+    if ( g_sinit == NULL )
         return TB_ERR_SINIT_NOT_PRESENT;
     /* do some checks on it */
-    if ( !verify_acmod(sinit) )
+    if ( !verify_acmod(g_sinit) )
         return TB_ERR_ACMOD_VERIFY_FAILED;
 
     /* print some debug info */
@@ -580,7 +704,7 @@ tb_error_t txt_launch_environment(const multiboot_info_t *mbi)
         return TB_ERR_FATAL;
 
     /* initialize TXT heap */
-    txt_heap = init_txt_heap(mle_ptab_base, sinit, mbi);
+    txt_heap = init_txt_heap(mle_ptab_base, g_sinit, lctx);
     if ( txt_heap == NULL )
         return TB_ERR_TXT_NOT_SUPPORTED;
 
@@ -589,43 +713,89 @@ tb_error_t txt_launch_environment(const multiboot_info_t *mbi)
     save_mtrrs(&(os_mle_data->saved_mtrr_state));
 
     /* set MTRRs properly for AC module (SINIT) */
-    if ( !set_mtrrs_for_acmod(sinit) )
+    if ( !set_mtrrs_for_acmod(g_sinit) )
         return TB_ERR_FATAL;
 
-    printk("executing GETSEC[SENTER]...\n");
+    printk(TBOOT_INFO"executing GETSEC[SENTER]...\n");
     /* (optionally) pause before executing GETSEC[SENTER] */
     if ( g_vga_delay > 0 )
         delay(g_vga_delay * 1000);
-    __getsec_senter((uint32_t)sinit, (sinit->size)*4);
-    printk("ERROR--we should not get here!\n");
+    __getsec_senter((uint32_t)g_sinit, (g_sinit->size)*4);
+    printk(TBOOT_INFO"ERROR--we should not get here!\n");
     return TB_ERR_FATAL;
 }
 
 bool txt_s3_launch_environment(void)
 {
-    acm_hdr_t *sinit;
-
     /* initial launch's TXT heap data is still in place and assumed valid */
     /* so don't re-create; this is OK because it was untrusted initially */
     /* and would be untrusted now */
 
+    /* initialize event log in os_sinit_data, so that events will not */
+    /* repeat when s3 */
+    if ( g_elog )
+        g_elog = (event_log_container_t *)init_event_log();
+
     /* get sinit binary loaded */
-    sinit = (acm_hdr_t *)(uint32_t)read_pub_config_reg(TXTCR_SINIT_BASE);
-    if ( sinit == NULL )
+    g_sinit = (acm_hdr_t *)(uint32_t)read_pub_config_reg(TXTCR_SINIT_BASE);
+    if ( g_sinit == NULL )
         return false;
 
     /* set MTRRs properly for AC module (SINIT) */
-    set_mtrrs_for_acmod(sinit);
+    set_mtrrs_for_acmod(g_sinit);
 
-    printk("executing GETSEC[SENTER]...\n");
+    printk(TBOOT_INFO"executing GETSEC[SENTER]...\n");
     /* (optionally) pause before executing GETSEC[SENTER] */
     if ( g_vga_delay > 0 )
         delay(g_vga_delay * 1000);
-    __getsec_senter((uint32_t)sinit, (sinit->size)*4);
-    printk("ERROR--we should not get here!\n");
+    __getsec_senter((uint32_t)g_sinit, (g_sinit->size)*4);
+    printk(TBOOT_ERR"ERROR--we should not get here!\n");
     return false;
 }
 
+tb_error_t txt_launch_racm(loader_ctx *lctx)
+{
+    acm_hdr_t *racm = NULL;
+
+    /*
+     * find correct revocation AC module in modules list
+     */
+    find_platform_racm(lctx, (void **)&racm, NULL);
+    /* copy it to a 32KB aligned memory address */
+    racm = copy_racm(racm);
+    if ( racm == NULL )
+        return TB_ERR_SINIT_NOT_PRESENT;
+    /* do some checks on it */
+    if ( !verify_racm(racm) )
+        return TB_ERR_ACMOD_VERIFY_FAILED;
+
+    /* save MTRRs before we alter them for RACM launch */
+    /*  - not needed by far since always reboot after RACM launch */
+    //save_mtrrs(...);
+
+    /* set MTRRs properly for AC module (RACM) */
+    if ( !set_mtrrs_for_acmod(racm) )
+        return TB_ERR_FATAL;
+
+    /* clear MSEG_BASE/SIZE registers */
+    write_pub_config_reg(TXTCR_MSEG_BASE, 0);
+    write_pub_config_reg(TXTCR_MSEG_SIZE, 0);
+
+    printk(TBOOT_INFO"executing GETSEC[ENTERACCS]...\n");
+    /* (optionally) pause before executing GETSEC[ENTERACCS] */
+    if ( g_vga_delay > 0 )
+        delay(g_vga_delay * 1000);
+    __getsec_enteraccs((uint32_t)racm, (racm->size)*4, 0xF0);
+    /* powercycle by writing 0x0a+0x0e to port 0xcf9, */
+    /* warm reset by write 0x06 to port 0xcf9 */
+    //outb(0xcf9, 0x0a);
+    //outb(0xcf9, 0x0e);
+    outb(0xcf9, 0x06);
+    
+    printk(TBOOT_ERR"ERROR--we should not get here!\n");
+    return TB_ERR_FATAL;
+}
+
 bool txt_prepare_cpu(void)
 {
     unsigned long eflags, cr0;
@@ -638,24 +808,24 @@ bool txt_prepare_cpu(void)
 
     /* must be in protected mode */
     if ( !(cr0 & CR0_PE) ) {
-        printk("ERR: not in protected mode\n");
+        printk(TBOOT_ERR"ERR: not in protected mode\n");
         return false;
     }
 
     /* cache must be enabled (CR0.CD = CR0.NW = 0) */
     if ( cr0 & CR0_CD ) {
-        printk("CR0.CD set\n");
+        printk(TBOOT_INFO"CR0.CD set\n");
         cr0 &= ~CR0_CD;
     }
     if ( cr0 & CR0_NW ) {
-        printk("CR0.NW set\n");
+        printk(TBOOT_INFO"CR0.NW set\n");
         cr0 &= ~CR0_NW;
     }
 
     /* native FPU error reporting must be enabled for proper */
     /* interaction behavior */
     if ( !(cr0 & CR0_NE) ) {
-        printk("CR0.NE not set\n");
+        printk(TBOOT_INFO"CR0.NE not set\n");
         cr0 |= CR0_NE;
     }
 
@@ -664,17 +834,17 @@ bool txt_prepare_cpu(void)
     /* cannot be in virtual-8086 mode (EFLAGS.VM=1) */
     eflags = read_eflags();
     if ( eflags & X86_EFLAGS_VM ) {
-        printk("EFLAGS.VM set\n");
+        printk(TBOOT_INFO"EFLAGS.VM set\n");
         write_eflags(eflags | ~X86_EFLAGS_VM);
     }
 
-    printk("CR0 and EFLAGS OK\n");
+    printk(TBOOT_INFO"CR0 and EFLAGS OK\n");
 
     /*
      * verify that we're not already in a protected environment
      */
     if ( txt_is_launched() ) {
-        printk("already in protected environment\n");
+        printk(TBOOT_ERR"already in protected environment\n");
         return false;
     }
 
@@ -686,13 +856,13 @@ bool txt_prepare_cpu(void)
     /* no machine check in progress (IA32_MCG_STATUS.MCIP=1) */
     mcg_stat = rdmsr(MSR_MCG_STATUS);
     if ( mcg_stat & 0x04 ) {
-        printk("machine check in progress\n");
+        printk(TBOOT_ERR"machine check in progress\n");
         return false;
     }
 
     getsec_parameters_t params;
     if ( !get_parameters(&params) ) {
-        printk("get_parameters() failed\n");
+        printk(TBOOT_ERR"get_parameters() failed\n");
         return false;
     }
 
@@ -701,22 +871,22 @@ bool txt_prepare_cpu(void)
     for ( unsigned int i = 0; i < (mcg_cap & 0xff); i++ ) {
         mcg_stat = rdmsr(MSR_MC0_STATUS + 4*i);
         if ( mcg_stat & (1ULL << 63) ) {
-            printk("MCG[%u] = %Lx ERROR\n", i, mcg_stat);
+            printk(TBOOT_ERR"MCG[%u] = %Lx ERROR\n", i, mcg_stat);
             if ( !params.preserve_mce )
                 return false;
         }
     }
 
     if ( params.preserve_mce )
-        printk("supports preserving machine check errors\n");
+        printk(TBOOT_INFO"supports preserving machine check errors\n");
     else
-        printk("no machine check errors\n");
+        printk(TBOOT_INFO"no machine check errors\n");
 
     if ( params.proc_based_scrtm )
-        printk("CPU support processor-based S-CRTM\n");
+        printk(TBOOT_INFO"CPU support processor-based S-CRTM\n");
 
     /* all is well with the processor state */
-    printk("CPU is ready for SENTER\n");
+    printk(TBOOT_INFO"CPU is ready for SENTER\n");
 
     return true;
 }
@@ -731,7 +901,7 @@ void txt_post_launch(void)
     err = txt_post_launch_verify_platform();
     /* don't return the error yet, because we need to restore settings */
     if ( err != TB_ERR_NONE )
-        printk("failed to verify platform\n");
+        printk(TBOOT_ERR"failed to verify platform\n");
 
     /* get saved OS state (os_mvmm_data_t) from LT heap */
     txt_heap = get_txt_heap();
@@ -748,7 +918,7 @@ void txt_post_launch(void)
     /* restore pre-SENTER IA32_MISC_ENABLE_MSR (no verification needed)
        (do after AP wakeup so that if restored MSR has MWAIT clear it won't
        prevent wakeup) */
-    printk("saved IA32_MISC_ENABLE = 0x%08x\n",
+    printk(TBOOT_DETA"saved IA32_MISC_ENABLE = 0x%08x\n",
            os_mle_data->saved_misc_enable_msr);
     wrmsr(MSR_IA32_MISC_ENABLE, os_mle_data->saved_misc_enable_msr);
     if ( use_mwait() ) {
@@ -768,18 +938,18 @@ void txt_post_launch(void)
     /* always set the TXT.CMD.SECRETS flag */
     write_priv_config_reg(TXTCR_CMD_SECRETS, 0x01);
     read_priv_config_reg(TXTCR_E2STS);   /* just a fence, so ignore return */
-    printk("set TXT.CMD.SECRETS flag\n");
+    printk(TBOOT_INFO"set TXT.CMD.SECRETS flag\n");
 
     /* open TPM locality 1 */
     write_priv_config_reg(TXTCR_CMD_OPEN_LOCALITY1, 0x01);
     read_priv_config_reg(TXTCR_E2STS);   /* just a fence, so ignore return */
-    printk("opened TPM locality 1\n");
+    printk(TBOOT_INFO"opened TPM locality 1\n");
 }
 
 void ap_wait(unsigned int cpuid)
 {
     if ( cpuid >= NR_CPUS ) {
-        printk("cpuid (%u) exceeds # supported CPUs\n", cpuid);
+        printk(TBOOT_ERR"cpuid (%u) exceeds # supported CPUs\n", cpuid);
         apply_policy(TB_ERR_FATAL);
         mtx_leave(&ap_lock);
         return;
@@ -792,11 +962,10 @@ void ap_wait(unsigned int cpuid)
     wrmsr(MSR_IA32_MISC_ENABLE, misc);
 
     /* this is close enough to entering monitor/mwait loop, so inc counter */
-    atomic_inc(&ap_wfs_count);
     atomic_inc((atomic_t *)&_tboot_shared.num_in_wfs);
     mtx_leave(&ap_lock);
 
-    printk("cpu %u mwait'ing\n", cpuid);
+    printk(TBOOT_INFO"cpu %u mwait'ing\n", cpuid);
     while ( _tboot_shared.ap_wake_trigger != cpuid ) {
         cpu_monitor(&_tboot_shared.ap_wake_trigger, 0, 0);
         mb();
@@ -818,14 +987,14 @@ void txt_cpu_wakeup(void)
     unsigned int cpuid = get_apicid();
 
     if ( cpuid >= NR_CPUS ) {
-        printk("cpuid (%u) exceeds # supported CPUs\n", cpuid);
+        printk(TBOOT_ERR"cpuid (%u) exceeds # supported CPUs\n", cpuid);
         apply_policy(TB_ERR_FATAL);
         return;
     }
 
     mtx_enter(&ap_lock);
 
-    printk("cpu %u waking up from TXT sleep\n", cpuid);
+    printk(TBOOT_INFO"cpu %u waking up from TXT sleep\n", cpuid);
 
     txt_heap = get_txt_heap();
     os_mle_data = get_os_mle_data_start(txt_heap);
@@ -840,9 +1009,10 @@ void txt_cpu_wakeup(void)
         apply_policy(TB_ERR_POST_LAUNCH_VERIFICATION);
 
     /* enable SMIs */
-    printk("enabling SMIs on cpu %u\n", cpuid);
+    printk(TBOOT_DETA"enabling SMIs on cpu %u\n", cpuid);
     __getsec_smctrl();
 
+    atomic_inc(&ap_wfs_count);
     if ( use_mwait() )
         ap_wait(cpuid);
     else
@@ -862,7 +1032,7 @@ tb_error_t txt_protect_mem_regions(void)
     /* TXT heap */
     base = read_pub_config_reg(TXTCR_HEAP_BASE);
     size = read_pub_config_reg(TXTCR_HEAP_SIZE);
-    printk("protecting TXT heap (%Lx - %Lx) in e820 table\n", base,
+    printk(TBOOT_INFO"protecting TXT heap (%Lx - %Lx) in e820 table\n", base,
            (base + size - 1));
     if ( !e820_protect_region(base, size, E820_RESERVED) )
         return TB_ERR_FATAL;
@@ -870,7 +1040,7 @@ tb_error_t txt_protect_mem_regions(void)
     /* SINIT */
     base = read_pub_config_reg(TXTCR_SINIT_BASE);
     size = read_pub_config_reg(TXTCR_SINIT_SIZE);
-    printk("protecting SINIT (%Lx - %Lx) in e820 table\n", base,
+    printk(TBOOT_INFO"protecting SINIT (%Lx - %Lx) in e820 table\n", base,
            (base + size - 1));
     if ( !e820_protect_region(base, size, E820_RESERVED) )
         return TB_ERR_FATAL;
@@ -878,7 +1048,8 @@ tb_error_t txt_protect_mem_regions(void)
     /* TXT private space */
     base = TXT_PRIV_CONFIG_REGS_BASE;
     size = TXT_CONFIG_REGS_SIZE;
-    printk("protecting TXT Private Space (%Lx - %Lx) in e820 table\n",
+    printk(TBOOT_INFO
+           "protecting TXT Private Space (%Lx - %Lx) in e820 table\n",
            base, (base + size - 1));
     if ( !e820_protect_region(base, size, E820_RESERVED) )
         return TB_ERR_FATAL;
@@ -891,12 +1062,12 @@ tb_error_t txt_protect_mem_regions(void)
     sinit_mdr_t *mdrs_base = (sinit_mdr_t *)(((void *)sinit_mle_data
                                               - sizeof(uint64_t)) +
                                              sinit_mle_data->mdrs_off);
-    printk("verifying e820 table against SINIT MDRs: ");
+    printk(TBOOT_INFO"verifying e820 table against SINIT MDRs: ");
     if ( !verify_e820_map(mdrs_base, num_mdrs) ) {
-        printk("verification failed.\n");
+        printk(TBOOT_ERR"verification failed.\n");
         return TB_ERR_POST_LAUNCH_VERIFICATION;
     }
-    printk("verification succeeded.\n");
+    printk(TBOOT_INFO"verification succeeded.\n");
 
     return TB_ERR_NONE;
 }
@@ -908,7 +1079,7 @@ void txt_shutdown(void)
     /* shutdown shouldn't be called on APs, but if it is then just hlt */
     apicbase = rdmsr(MSR_APICBASE);
     if ( !(apicbase & APICBASE_BSP) ) {
-        printk("calling txt_shutdown on AP\n");
+        printk(TBOOT_INFO"calling txt_shutdown on AP\n");
         while ( true )
             halt();
     }
@@ -916,17 +1087,18 @@ void txt_shutdown(void)
     /* set TXT.CMD.NO-SECRETS flag (i.e. clear SECRETS flag) */
     write_priv_config_reg(TXTCR_CMD_NO_SECRETS, 0x01);
     read_priv_config_reg(TXTCR_E2STS);   /* fence */
-    printk("secrets flag cleared\n");
+    printk(TBOOT_INFO"secrets flag cleared\n");
 
     /* unlock memory configuration */
     write_priv_config_reg(TXTCR_CMD_UNLOCK_MEM_CONFIG, 0x01);
     read_pub_config_reg(TXTCR_E2STS);    /* fence */
-    printk("memory configuration unlocked\n");
+    printk(TBOOT_INFO"memory configuration unlocked\n");
 
     /* if some APs are still in wait-for-sipi then SEXIT will hang */
-    /* so TXT reset the platform instead */
-    if ( atomic_read(&ap_wfs_count) > 0 ) {
-        printk("exiting with some APs still in wait-for-sipi state (%u)\n",
+    /* so TXT reset the platform instead, expect mwait case */
+    if ( (!use_mwait()) && atomic_read(&ap_wfs_count) > 0 ) {
+        printk(TBOOT_INFO
+               "exiting with some APs still in wait-for-sipi state (%u)\n",
                atomic_read(&ap_wfs_count));
         write_priv_config_reg(TXTCR_CMD_RESET, 0x01);
     }
@@ -936,15 +1108,15 @@ void txt_shutdown(void)
     read_priv_config_reg(TXTCR_E2STS);   /* fence */
     write_priv_config_reg(TXTCR_CMD_CLOSE_PRIVATE, 0x01);
     read_pub_config_reg(TXTCR_E2STS);    /* fence */
-    printk("private config space closed\n");
+    printk(TBOOT_INFO"private config space closed\n");
 
     /* SMXE may not be enabled any more, so set it to make sure */
     write_cr4(read_cr4() | CR4_SMXE);
 
     /* call GETSEC[SEXIT] */
-    printk("executing GETSEC[SEXIT]...\n");
+    printk(TBOOT_INFO"executing GETSEC[SEXIT]...\n");
     __getsec_sexit();
-    printk("measured environment torn down\n");
+    printk(TBOOT_INFO"measured environment torn down\n");
 }
 
 bool txt_is_powercycle_required(void)
@@ -975,7 +1147,7 @@ bool get_parameters(getsec_parameters_t *params)
     /* sanity check because GETSEC[PARAMETERS] will fail if not set */
     cr4 = read_cr4();
     if ( !(cr4 & CR4_SMXE) ) {
-        printk("SMXE not enabled, can't read parameters\n");
+        printk(TBOOT_ERR"SMXE not enabled, can't read parameters\n");
         return false;
     }
 
@@ -998,7 +1170,7 @@ bool get_parameters(getsec_parameters_t *params)
         /* supported ACM versions */
         else if ( param_type == 1 ) {
             if ( params->n_versions == MAX_SUPPORTED_ACM_VERSIONS )
-                printk("number of supported ACM version exceeds "
+                printk(TBOOT_WARN"number of supported ACM version exceeds "
                        "MAX_SUPPORTED_ACM_VERSIONS\n");
             else {
                 params->acm_versions[params->n_versions].mask = ebx;
@@ -1021,7 +1193,8 @@ bool get_parameters(getsec_parameters_t *params)
             params->preserve_mce = (eax & 0x00000040) ? true : false;
         }
         else {
-            printk("unknown GETSEC[PARAMETERS] type: %d\n", param_type);
+            printk(TBOOT_WARN"unknown GETSEC[PARAMETERS] type: %d\n", 
+                   param_type);
             param_type = 0;    /* set so that we break out of the loop */
         }
     } while ( param_type != 0 );
